"""Tests for ObjectRetrieval() class."""

import xml.etree.ElementTree as ET

from io import BytesIO
from PIL import Image

from pybliometrics.sciencedirect import init, ObjectRetrieval

init()

or_1 = ObjectRetrieval('S156984322300331X',
                       'gr10.jpg',
                       refresh=30)
or_2 = ObjectRetrieval('10.1016/j.rcim.2020.102086',
                       'si92.svg',
                       id_type='doi',
                       refresh=30)


def test_object():
    """Tests whether the object is a BytesIO object and its content."""
    obj_1_last_50 = b'\xbf\xbd\xb6\xeb;+\\\x87Y7\x94[y\x17\xe3\xeb/(\xcf#\xda\xc9\x90\x80 \x08\x02\x00\x80 \x08\x02\x00\x80 \x08\x02\x00\x80 \x08\x02\x00\x80 \x08\x02\x03\xff\xd9'
    assert isinstance(or_1.object, BytesIO)
    assert or_1.object.getvalue()[-50:] == obj_1_last_50
    with Image.open(or_1.object) as img:
        assert img.format.lower() == 'jpeg'

    obj_2_150_200 = b"085 235.866 8.8237' width='283.039pt' xmlns='http:"
    assert isinstance(or_2.object, BytesIO)
    assert or_2.object.getvalue()[150:200] == obj_2_150_200
    assert ET.parse(or_2.object).getroot().tag == '{http://www.w3.org/2000/svg}svg'

def test_str():
    """Tests the string representation of the ObjectRetrieval object."""
    expected_1 = "Object gr10.jpg from document with EID 1-s2.0-S156984322300331X has size of 34.7 KB."
    assert str(or_1) == expected_1
    expected_2 = "Object si92.svg from document with EID 1-s2.0-S0736584520302969 has size of 10.0 KB."
    assert str(or_2) == expected_2



from typing import NamedTuple

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import check_parameter_value


class Category(NamedTuple):
    name: str
    total: int


class Metric(NamedTuple):
    name: str
    total: int


class PlumXMetrics(Retrieval):
    @property
    def category_totals(self) -> list[Category] | None:
        """A list of namedtuples representing total metrics as categorized
        by PlumX Metrics in the form `(capture, citation, mention, socialMedia,
        usage)`.

        Note: For Citation category a maximum citation count across sources
        is shown.  For details on PlumX Metrics categories see
        https://plumanalytics.com/learn/about-metrics/.
        """
        categories = self._json.get('count_categories', [])
        return _format_as_category_list(categories) or None

    @property
    def capture(self) -> list[Metric] | None:
        """A list of namedtuples representing metrics in the Captures category.

        Note: For details on Capture metrics see
        https://plumanalytics.com/learn/about-metrics/capture-metrics/.
        """
        metrics = self._count_categories.get('capture', [])
        return _format_as_metric_list(metrics) or None

    @property
    def citation(self) -> list[Metric] | None:
        """A list of namedtuples representing citation counts from
        different sources.

        Note: For details on Citation metrics see
        https://plumanalytics.com/learn/about-metrics/citation-metrics/.
        """
        metrics = []
        for item in self._count_categories.get('citation', []):
            if item.get('sources'):
                metrics += item['sources']
        return _format_as_metric_list(metrics) or None

    @property
    def mention(self) -> list[Metric] | None:
        """A list of namedtuples representing metrics in Mentions category.

        Note: For details on Mention metrics see
        https://plumanalytics.com/learn/about-metrics/mention-metrics/.
        """
        metrics = self._count_categories.get('mention', [])
        return _format_as_metric_list(metrics) or None

    @property
    def social_media(self) -> list[Metric] | None:
        """A list of namedtuples representing social media metrics.

        Note: For details on Social Media metrics see
        https://plumanalytics.com/learn/about-metrics/social-media-metrics/.
        """
        metrics = self._count_categories.get('socialMedia', [])
        return _format_as_metric_list(metrics) or None

    @property
    def usage(self) -> list[Metric] | None:
        """A list of namedtuples representing Usage category metrics.

        Note: For details on Usage metrics see
        https://plumanalytics.com/learn/about-metrics/usage-metrics/.
        """
        metrics = self._count_categories.get('usage', [])
        return _format_as_metric_list(metrics) or None

    def __init__(self,
                 identifier: str,
                 id_type: str,
                 refresh: bool | int = False,
                 **kwds: str
                 ) -> None:
        """Interaction with the PlumX Metrics API.

        :param identifier: The identifier of a document.
        :param id_type: The type of used ID. Allowed values are:
                        'airitiDocId'; 'cabiAbstractId'; 'citeulikeId';
                        'digitalMeasuresArtifactId'; 'doi'; 'elsevierId';
                        'elsevierPii'; 'facebookCountUrlId';
                        'figshareArticleId'; 'isbn'; 'lccn'; 'medwaveId';
                        'nctId'; 'oclc'; 'pittEprintDscholarId'; 'pmcid';
                        'pmid'; 'redditId'; 'repecHandle'; 'repoUrl';
                        'scieloId'; 'sdEid'; 'slideshareUrlId';
                        'smithsonianPddrId'; 'ssrnId'; 'urlId'
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If `int` is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/PlumXMetricsAPI.wadl.

        Raises
        ------
        ValueError
            If the parameter `refresh` is not one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/ENHANCED/{identifier}`,
        where `path` is specified in your configuration file.
        """
        # Checks
        allowed = ('airitiDocId', 'cabiAbstractId', 'citeulikeId',
                   'digitalMeasuresArtifactId', 'doi', 'elsevierId',
                   'elsevierPii', 'facebookCountUrlId', 'figshareArticleId',
                   'isbn', 'lccn', 'medwaveId', 'nctId', 'oclc',
                   'pittEprintDscholarId', 'pmcid', 'pmid', 'redditId',
                   'repecHandle', 'repoUrl', 'scieloId', 'sdEid',
                   'slideshareUrlId', 'smithsonianPddrId', 'ssrnId', 'urlId')
        check_parameter_value(id_type, allowed, "id_type")
        self._id_type = id_type
        self._identifier = identifier

        # Load json
        self._refresh = refresh
        self._view = 'ENHANCED'
        Retrieval.__init__(self, identifier=identifier, id_type=id_type, **kwds)
        cats = self._json.get('count_categories', [])
        self._count_categories = {d["name"]: d['count_types'] for d in cats}

    def __str__(self):
        """Print a summary string."""
        s = f"Document with {self._id_type} {self._identifier} received:\n- "
        cats = [f"{c.total:,} citation(s) in category '{c.name}'"
                for c in self.category_totals]
        s += "\n- ".join(cats)
        s += f"\nas of {self.get_cache_file_mdate().split()[0]}"
        return s


def _format_as_category_list(metric_counts):
    """Formats list of dicts of metrics into list of Category namedtuples."""
    return [Category(name=t['name'], total=t['total']) for t in metric_counts]


def _format_as_metric_list(metric_counts):
    """Formats list of dicts of metrics into list of Metric namedtuples."""
    return [Metric(name=t['name'], total=t['total']) for t in metric_counts]



import configparser
from pathlib import Path
from pybliometrics.utils.constants import CONFIG_FILE


def create_config(config_dir: str | Path | None = None,
                  keys: list[str] | None = None,
                  insttoken: list[str] | None = None
                  ):
    """Initiates process to generate configuration file.

    :param config_dir: The location of the configuration file.
    :param keys: If you provide a list of keys, pybliometrics will skip the
                 prompt.  It will also not ask for InstToken.  This is
                 intended for workflows using CI, not for general use.
    :param insttoken: An InstToken to be used alongside the key(s). Will only
                      be used if `keys` is not empty.
    """
    from pybliometrics.utils.constants import DEFAULT_PATHS

    if not config_dir:
        config_dir = CONFIG_FILE

    config = configparser.ConfigParser()
    config.optionxform = str
    print(f"Creating config file at {config_dir} with default paths...")

    # Set directories
    config.add_section('Directories')
    for api, path in DEFAULT_PATHS.items():
        config.set('Directories', api, str(path))

    # Set authentication
    config.add_section('Authentication')

    # Get keys and tokens
    if keys:
        if not isinstance(keys, list):
            raise ValueError("Parameter `keys` must be a list.")
        keys = ", ".join(keys)
    if insttoken:
        if not isinstance(insttoken, list):
            raise ValueError("Parameter `inst_tokens` must be a list. ")
        insttoken = ", ".join(insttoken)

    # If no keys or tokens are provided, ask for them
    if not (keys or insttoken):
        prompt_key = "Please enter your API Key(s), obtained from "\
                     "http://dev.elsevier.com/myapikey.html.  Separate "\
                     "multiple keys by comma:\n"
        keys = input(prompt_key)
        prompt_token = "API Keys are sufficient for most users.  If you "\
                       "have an InstToken, please enter the tokens pair now. "\
                       "Separate multiple tokens by a comma. The correspondig "\
                       "key's position should match the position of the token."\
                       "If you don't have tokens, just press Enter:\n"
        insttoken = input(prompt_token)

    # Set keys and tokens in config
    config.set('Authentication', 'APIKey', keys)
    if insttoken:
        config.set('Authentication', 'InstToken', insttoken)

    # Set default values
    config.add_section('Requests')
    config.set('Requests', 'Timeout', '20')
    config.set('Requests', 'Retries', '5')

    # Write out
    config_dir.parent.mkdir(parents=True, exist_ok=True)
    with open(config_dir, "w") as ouf:
        config.write(ouf)
    print(f"Configuration file successfully created at {config_dir}\n"
          "For details see https://pybliometrics.rtfd.io/en/stable/configuration.html.")
    return config



from datetime import datetime
from hashlib import md5
from typing import NamedTuple
from warnings import warn

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import chained_get, check_parameter_value


class Author(NamedTuple):
    name: str | None
    surname: str | None
    initials: str | None
    id: str
    url: str | None


class CitationOverview(Retrieval):
    @property
    def authors(self) -> list[list[Author] | None] | None:
        """A list of lists of namedtuples storing author information,
        where each namedtuple corresponds to one author and each sub-list to
        one document.
        The information in each namedtuple is `(name surname initials id url)`.
        All entries are strings.
        """
        outer = []
        for doc in self._citeInfoMatrix:
            inner = []
            for author in doc.get('author', []):
                author = {k.split(":", 1)[-1]: v for k, v in author.items()}
                new = Author(name=author.get('index-name'),
                             id=author['authid'],
                             surname=author.get('surname'),
                             initials=author.get('initials'),
                             url=author.get('author-url'))
                inner.append(new)
            outer.append(inner or None)
        return _maybe_return_list(outer)

    @property
    def cc(self) -> list[list[tuple[int, int]]]:
        """List of lists of tuples of yearly number of citations for specified
        years, where each sub-list corresponds to one document.
        """
        try:
            dates = self._date.split("-")
        except AttributeError:
            current_year = datetime.now().year
            dates = (current_year-2, current_year)
        _years = range(int(dates[0]), int(dates[1])+1)
        outer = []
        for doc in self._citeInfoMatrix:
            try:
                cites = [int(d['$']) for d in doc['cc']]
            except (AttributeError, KeyError, TypeError):  # No citations
                cites = [0]*len(_years)
            outer.append(list(zip(_years, cites)))
        return _maybe_return_list(outer)

    @property
    def citationType_long(self) -> list[str | None] | None:
        """Type (long version) of the documents (e.g. article, review)."""
        path = ["citationType", "$"]
        out = [chained_get(e, path) for e in self._citeInfoMatrix]
        return _maybe_return_list(out)

    @property
    def citationType_short(self) -> list[str | None] | None:
        """Type (short version) of the documents (e.g. ar, re)."""
        path = ["citationType", "@code"]
        out = [chained_get(e, path) for e in self._citeInfoMatrix]
        return _maybe_return_list(out)

    @property
    def columnTotal(self) -> list[int]:
        """The yearly number of citations for all documents combined."""
        return [int(d["$"]) for d in self._citeCountHeader["columnTotal"]]

    @property
    def doi(self) -> list[str | None] | None:
        """Document Object Identifier (DOI) of the documents."""
        out = [e.get('doi') for e in self._identifierlegend]
        return _maybe_return_list(out)

    @property
    def endingPage(self) -> list[str | None] | None:
        """Ending pages of the documents."""
        out = [e.get('endingPage') for e in self._citeInfoMatrix]
        return _maybe_return_list(out)

    @property
    def grandTotal(self) -> int:
        """The total number of citations of all documents together."""
        return int(self._citeCountHeader["grandTotal"])

    @property
    def h_index(self) -> int:
        """Combined h-index of citations of all the documents."""
        return int(self._data['h-index'])

    @property
    def issn(self) -> list[str | tuple[str, str] | None] | None:
        """ISSN of the publishers of the documents.
        Note: If E-ISSN is known to Scopus, this returns both
        ISSN and E-ISSN in random order separated by blank space.
        """
        out = [e.get('issn') for e in self._citeInfoMatrix]
        return _maybe_return_list(out)

    @property
    def issueIdentifier(self) -> list[str | None] | None:
        """Issue numbers of the documents."""
        out = [e.get('issueIdentifier') for e in self._citeInfoMatrix]
        return _maybe_return_list(out)

    @property
    def laterColumnTotal(self) -> int:
        """The total number of citations for all years after the end
        year for all documents combined.
        """
        return int(self._citeCountHeader["laterColumnTotal"])

    @property
    def lcc(self) -> list[int]:
        """Number of citations after the end year of each document."""
        return [int(m['lcc']) for m in self._citeInfoMatrix]

    @property
    def pcc(self) -> list[int]:
        """Number of citations before the start year."""
        return [int(m['pcc']) for m in self._citeInfoMatrix]

    @property
    def pii(self) -> list[str | None] | None:
        """The Publication Item Identifier (PII) of the documents."""
        out = [e.get('pii') for e in self._identifierlegend]
        return _maybe_return_list(out)

    @property
    def prevColumnTotal(self) -> int:
        """The total number of citations for all years before the start
        year for all documents combined.
        """
        return int(self._citeCountHeader["prevColumnTotal"])

    @property
    def publicationName(self) -> list[str | None] | None:
        """Name of source the documents are published in (e.g. the Journal)."""
        out = [e.get('publicationName') for e in self._citeInfoMatrix]
        return _maybe_return_list(out)

    @property
    def rangeColumnTotal(self) -> int:
        """The total number of citations for all specified years for all
        documents combined.
        """
        return int(self._citeCountHeader["rangeColumnTotal"])

    @property
    def rangeCount(self) -> list[int]:
        """Total citation count over the specified year range for
        each document.
        """
        return [int(e['rangeCount']) for e in self._citeInfoMatrix]

    @property
    def rowTotal(self) -> list[int]:
        """Total number of citations (specified and omitted years) for each
        document.
        """
        return [int(e['rowTotal']) for e in self._citeInfoMatrix]

    @property
    def scopus_id(self) -> list[int]:
        """The Scopus ID(s) of the documents.  Might differ from the
        ones provided.
        """
        return [int(e['scopus_id']) for e in self._identifierlegend]

    @property
    def startingPage(self) -> list[str | None] | None:
        """Starting page."""
        out = [e.get('startingPage') for e in self._citeInfoMatrix]
        return _maybe_return_list(out)

    @property
    def title(self) -> list[str]:
        """Titles of each document."""
        return [e["title"] for e in self._citeInfoMatrix]

    @property
    def url(self) -> list[str]:
        """URL(s) to Citation Overview API view of each document."""
        return [e["url"] for e in self._citeInfoMatrix]

    @property
    def volume(self) -> list[str | None] | None:
        """Volume for the abstract."""
        out = [e.get('volume') for e in self._citeInfoMatrix]
        return _maybe_return_list(out)

    def __init__(self,
                 identifier: list[int | str],
                 date: str | None = None,
                 start: int | str | None = None,
                 end: int | str | None = None,
                 id_type: str = "scopus_id",
                 refresh: bool | int = False,
                 citation: str | None = None,
                 **kwds: str
                 ) -> None:
        """Interaction with the Citation Overview API.

        :param identifier: Up to 25 identifiers of the same kind for which to look
                           up citations.  Must be Scopus IDs, DOIs, PIIs or
                           Pubmed IDs.
        :param date: Represents the year range for which the citations should be counted.
                     If `None`, Scopus returns data for the current and the previous
                     two years.
        :param start: (deprecated) The first year for which the citation count should
                      be loaded.
        :param end: (deprecated) The last year for which the citation count should be
                    loaded. Defaults to the current year.
        :param id_type: The type of the IDs provided in `identifier`.  Must be
                        one of `"scopus_id", "doi", "pii", "pubmed_id"`.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param citation: Allows for the exclusion of self-citations or those
                         by books.  If `None`, will count all citations.
                         Allowed values: `None, exclude-self, exclude-books`
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/AbstractCitationAPI.wadl.

        Raises
        -----
        ValueError
            If parameter `identifier` contains fewer than 1 or more than
            25 elements.

        ValueError
            If any of the parameters `citation`, `id_type` or `refresh` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/STANDARD/{id}-{citation}-{date}`,
        where `path` is specified in your configuration file, and `id` the
        md5-hashed version of a string joining `identifier` on underscore.

        Your API Key needs to be augmented by Elsevier's Scopus
        Integration Team to access this API.
        """
        # Checks
        allowed = ('scopus_id', 'doi', 'pii', 'pubmed_id')
        check_parameter_value(id_type, allowed, "id_type")
        if citation:
            allowed = ('exclude-self', 'exclude-books')
            check_parameter_value(citation, allowed, "citation")
        if len(identifier) < 0 or len(identifier) > 25:
            msg = "Provide at least 1 and at most than 25 identifiers"
            raise ValueError(msg)

        # Variables
        identifier = [str(i) for i in identifier]
        if start or end:
            msg = "Parameters `start` and `end` are deprecated and will be removed"\
                  f" in a future release.  Please use 'date={start}-{end}' instead."
            warn(msg, FutureWarning)
            if not date:
                date = f'{start}-{end}'
        self._date = date
        self._citation = citation
        self._refresh = refresh
        self._view = "STANDARD"

        # Get file content
        kwds.update({id_type: identifier})
        stem = md5("_".join(identifier).encode('utf8')).hexdigest()
        Retrieval.__init__(self, stem, date=date, citation=citation, **kwds)
        self._data = self._json['abstract-citations-response']

        # citeInfoMatrix
        matrix = self._data['citeInfoMatrix']['citeInfoMatrixXML']['citationMatrix']['citeInfo']
        self._citeInfoMatrix = [_parse_dict(e) for e in matrix]
        # identifier-legend
        identifier = self._data['identifier-legend']['identifier']
        self._identifierlegend = [_parse_dict(e) for e in identifier]
        # citeCountHeader
        self._citeCountHeader = self._data['citeColumnTotalXML']["citeCountHeader"]

    def __str__(self):
        """Return a summary string."""
        cits_dict = {'exclude-self': 'excluding self-citations',
                     'exclude-books': 'excluding citations from books'}
        date = self.get_cache_file_mdate().split()[0]
        cits_type = f'{cits_dict.get(self._citation, "")}'
        s = f"{len(self.scopus_id)} document(s) has/have the following "\
            f"total citation count{cits_type} as of {date}:\n    "\
            f"{'; '.join([str(n) for n in self.rowTotal])}"
        return s


def _parse_dict(dct):
    """Auxiliary function to change the keys of a dictionary."""
    return {k.split(":", 1)[-1]: v for k, v in dct.items()}


def _maybe_return_list(lst):
    """Return `lst` unless all of its elements are empty."""
    if all(e is None for e in lst):
        return None
    else:
        return lst



"""Tests for sciencedirect.ScienceDirectSearch"""

from pybliometrics.exception import Scopus400Error
from pybliometrics.sciencedirect import ScienceDirectSearch, init
from pybliometrics.sciencedirect.sciencedirect_search import Document

init()

sds_standard = ScienceDirectSearch('TITLE("Assessing LLMs in malicious code deobfuscation of real-world malware campaigns") AND DATE(2024)', view="STANDARD", refresh=30)
sds_empty = ScienceDirectSearch('TITLE("Not a very realistic title")', view="STANDARD", refresh=30)


def test_empty_results():
    assert sds_empty.results is None
    assert sds_empty._n == 0


def test_all_fields():
    expected_standard_doc = Document(
        authors="Constantinos Patsakis;Fran Casino;Nikolaos Lykousas",
        first_author="Constantinos Patsakis",
        doi="10.1016/j.eswa.2024.124912",
        title="Assessing LLMs in malicious code deobfuscation of real-world malware campaigns",
        link="https://www.sciencedirect.com/science/article/pii/S0957417424017792?dgcid=api_sd_search-api-endpoint",
        load_date="2024-07-31T00:00:00.000Z",
        openaccess_status=True,
        pii="S0957417424017792",
        coverDate="2024-12-05",
        endingPage=None,
        publicationName="Expert Systems with Applications",
        startingPage="124912",
        api_link="https://api.elsevier.com/content/article/pii/S0957417424017792",
        volume="256",
    )
    assert sds_standard.results[0] == expected_standard_doc


def test_field_consistency():
    am_wrong_field = ScienceDirectSearch('TITLE("Assessing LLMs in malicious code deobfuscation of real-world malware campaigns") AND DATE(2012)',
                                 integrity_fields=["notExistingField"],
                                 integrity_action="warn",
                                 view="STANDARD",
                                 refresh=30)
    try:
        am_wrong_field.results
    except ValueError:
        pass
    except Exception as e:
        raise AssertionError(f"Unexpected exception type: {type(e).__name__}")
    else:
        raise AssertionError("Expected ValueError but no exception was raised")


def test_length():
    assert len(sds_standard.results) == sds_standard._n
    assert len(sds_standard.results) == sds_standard._n


def test_string():
    str_start = ('Search \'TITLE("Assessing LLMs in malicious code deobfuscation of '
    'real-world malware campaigns") AND DATE(2024)\' yielded 1 document as of')
    assert sds_standard.__str__().startswith(str_start)


def test_wrong_query():
    try:
        ScienceDirectSearch(
            'Th(s querY - has M&ny ( Errors', view="STANDARD", refresh=30
        )
    except Scopus400Error:
        pass
    except Exception as e:
        raise AssertionError(f"Unexpected exception type: {type(e).__name__}")
    else:
        raise AssertionError("Expected Scopus400Error but no exception was raised")



"""Tests for ObjectMetadata() class."""

from pybliometrics.sciencedirect import init, ObjectMetadata
from pybliometrics.sciencedirect.object_metadata import Metadata

init()

om_1 = ObjectMetadata('10.1016/j.neunet.2024.106632', refresh=30)
om_2 = ObjectMetadata('S2213305418300365', id_type='pii', refresh=30)


def test_results():
    """Tests the length and fields of `results`."""
    fields = ('eid', 'filename', 'height', 'mimetype', 'ref', 'size', 'type', 'url', 'width')

    assert om_1.results[0]._fields == fields
    assert len(om_1.results) == 355

    expected_om_1 = Metadata(eid='1-s2.0-S0893608024005562-gr3.jpg',
                         filename='gr3.jpg',
                         height=729,
                         mimetype='image/jpeg',
                         ref='gr3',
                         size=100202,
                         type='IMAGE-DOWNSAMPLED',
                         url='https://api.elsevier.com/content/object/eid/1-s2.0-S0893608024005562-gr3.jpg?httpAccept=%2A%2F%2A',
                         width=656)

    assert om_1.results[0] == expected_om_1

    assert om_2.results[2]._fields == fields
    assert len(om_2.results) == 18

    expected_om_2 = Metadata(eid='1-s2.0-S2213305418300365-gr2.sml',
                             filename='gr2.sml',
                             height=146,
                             mimetype='image/gif',
                             ref='gr2',
                             size=14910,
                             type='IMAGE-THUMBNAIL',
                             url='https://api.elsevier.com/content/object/eid/1-s2.0-S2213305418300365-gr2.sml?httpAccept=%2A%2F%2A',
                             width=219)

    assert om_2.results[2] == expected_om_2



from warnings import warn
from typing import NamedTuple

from json import loads

from .author_search import AuthorSearch
from .scopus_search import ScopusSearch
from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import chained_get, check_parameter_value,\
    filter_digits, get_content, get_link, html_unescape, listify, make_int_if_possible,\
    parse_affiliation, parse_date_created, VIEWS


class Affiliation(NamedTuple):
    id: int | None
    parent: int | None
    type: str | None
    relationship: str | None
    afdispname: str | None
    preferred_name: str | None
    parent_preferred_name: str | None
    country_code: str | None
    country: str | None
    address_part: str | None
    city: str | None
    state: str | None
    postal_code: str | None
    org_domain: str | None
    org_URL: str | None


class Variant(NamedTuple):
    indexed_name: str | None
    initials: str | None
    surname: str | None
    given_name: str | None
    doc_count: int | None


class Subjectarea(NamedTuple):
    area: str
    abbreviation: str
    code: int


class Coauthor(NamedTuple):
    surname: str
    given_name: str | None
    id: int
    areas: str
    affiliation_id: str | None
    name: str | None
    city: str | None
    country: str | None


class AuthorRetrieval(Retrieval):
    @property
    def affiliation_current(self) -> list[Affiliation] | None:
        """A list of namedtuples representing the authors's current
        affiliation(s), in the form `(id parent type relationship afdispname
        preferred_name parent_preferred_name country_code country address_part
        city state postal_code org_domain org_URL)`.
        Note: Affiliation information might be missing or mal-assigned even
        when it lookes correct in the web view.  In this case please request
        a correction.
        """
        if self._view in ('STANDARD', 'ENHANCED'):
            affs = chained_get(self._profile, ["affiliation-current", "affiliation"])
        elif self._view == 'LIGHT':
            affs = self._json.get('affiliation-current')
        else:
            return None
        return parse_affiliation(affs or {}, self._view)

    @property
    def affiliation_history(self) -> list[Affiliation] | None:
        """A list of namedtuples representing the authors's historical
        affiliation(s), in the form `(id parent type relationship afdispname
        preferred_name parent_preferred_name country_code country address_part
        city state postal_code org_domain org_URL)`.
        Note: Affiliation information might be missing or mal-assigned even
        when it lookes correct in the web view.  In this case please request
        a correction.

        Note: Unlike on their website, Scopus doesn't provide the periods
        of affiliation.
        """
        affs = chained_get(self._profile, ["affiliation-history", "affiliation"])
        return parse_affiliation(affs or {}, self._view)

    @property
    def alias(self) -> list[str] | None:
        """List of possible new Scopus Author Profile IDs in case the profile
        has been merged.
        """
        return self._alias

    @property
    def citation_count(self) -> int:
        """Total number of citing items."""
        return make_int_if_possible(chained_get(self._json, ['coredata', 'citation-count']))

    @property
    def cited_by_count(self) -> int:
        """Total number of citing authors."""
        return make_int_if_possible(chained_get(self._json, ['coredata', 'cited-by-count']))

    @property
    def classificationgroup(self) -> list[tuple[int, int]] | None:
        """List with tuples with form`(subject group ID, number of documents)`."""
        path = ['classificationgroup', 'classifications', 'classification']
        out = [(int(filter_digits(item['$'])), int(filter_digits(item['@frequency'])))
               for item in listify(chained_get(self._profile, path, []))]
        return out or None

    @property
    def coauthor_count(self) -> int | None:
        """Total number of coauthors."""
        return make_int_if_possible(chained_get(self._json, ['coauthor-count']))

    @property
    def coauthor_link(self) -> str | None:
        """URL to Scopus API search page for coauthors."""
        return get_link(self._json, 3)

    @property
    def date_created(self) -> tuple[int, int, int] | None:
        """Date the Scopus record was created."""
        try:
            return parse_date_created(self._profile)
        except KeyError:
            return None

    @property
    def document_count(self) -> int:
        """Number of documents authored (excludes book chapters and notes)."""
        return make_int_if_possible(chained_get(self._json, ['coredata', 'document-count']))
    
    @property
    def document_entitlement_status(self) -> str | None:
        """Returns the document entitlement status, i.e. tells if the requestor 
        is entitled to the requested resource.
        Note: Only works with `ENTITLED` view.
        """
        return chained_get(self._json, ['document-entitlement', 'status'])

    @property
    def eid(self) -> str | None:
        """The EID of the author.  If it differs from the one provided,
        pybliometrics will throw a warning informing the user about
        author profile merges.
        """
        return chained_get(self._json, ['coredata', 'eid'])

    @property
    def given_name(self) -> str | None:
        """Author's preferred given name."""
        return html_unescape(chained_get(self._profile, ['preferred-name', 'given-name']))

    @property
    def h_index(self) -> int | None:
        """The author's h-index."""
        return make_int_if_possible(chained_get(self._json, ['h-index']))

    @property
    def historical_identifier(self) -> list[int] | None:
        """Scopus IDs of previous profiles now compromising this profile."""
        hist = chained_get(self._json, ["coredata", 'historical-identifier'], [])
        return [int(d['$'].split(":")[-1]) for d in hist] or None

    @property
    def identifier(self) -> int:
        """The author's ID.  Might differ from the one provided."""
        ident = chained_get(self._json, ['coredata', 'dc:identifier'])
        if not ident:
            return ident
        ident = ident.split(":")[-1]
        if ident != self._id:
            text = f"Profile with ID {self._id} has been merged and the new "\
                   f"ID is {ident}.  Please update your records manually.  "\
                   "Files have been cached with the old ID."
            warn(text, UserWarning)
        return int(ident)

    @property
    def indexed_name(self) -> str | None:
        """Author's name as indexed by Scopus."""
        if self._view in ('STANDARD', 'ENHANCED'):
            indexed_name = html_unescape(chained_get(self._profile, ['preferred-name', 'indexed-name']))
        elif self._view == 'LIGHT':
            # Try to get indexed name from name-variants
            name_variants = chained_get(self._json, ['name-variants', 'name-variant'])
            if name_variants:
                indexed_name = chained_get(name_variants[0], ['name-variant', 'indexed-name'])
            else:
                # In case of no name-variants get name from preferred-name
                preferred_name = self._json.get('preferred-name')
                indexed_name = ' '.join([preferred_name.get('initials', ''), preferred_name.get('surname', '')])
        else:
            indexed_name = None
        
        return indexed_name

    @property
    def initials(self) -> str | None:
        """Author's preferred initials."""
        return html_unescape(chained_get(self._profile, ['preferred-name', 'initials']))

    @property
    def name_variants(self) -> list[Variant] | None:
        """List of named tuples containing variants of the author name with
        number of documents published with that variant.
        """
        out = [Variant(indexed_name=html_unescape(var['indexed-name']), surname=html_unescape(var['surname']),
                       doc_count=make_int_if_possible(var.get('@doc-count')),
                       initials=html_unescape(var['initials']),
                       given_name=html_unescape(var.get('given-name')))
               for var in listify(self._profile.get('name-variant', []))]
        return out or None

    @property
    def orcid(self) -> str | None:
        """The author's ORCID."""
        return chained_get(self._json, ['coredata', 'orcid'])

    @property
    def publication_range(self) -> tuple[int, int] | None:
        """Tuple containing years of first and last publication."""        
        if self._view in ('STANDARD', 'ENHANCED', 'LIGHT'):
            if self._view in ('STANDARD', 'ENHANCED'):
                r = self._profile.get('publication-range')
                start = '@start'
                end = '@end'
            elif self._view == 'LIGHT':
                r = self._json.get('publication-range')
                start = 'start'
                end = 'end'
            
            try:
                return int(r.get(start)), int(r.get(end))
            except TypeError:
                return None
            
        return None

    @property
    def scopus_author_link(self) -> str | None:
        """Link to the Scopus web view of the author."""
        return get_link(self._json, 1)

    @property
    def search_link(self) -> str | None:
        """URL to the API page listing documents of the author."""
        return get_link(self._json, 2)

    @property
    def self_link(self) -> str | None:
        """Link to the author's API page."""
        return get_link(self._json, 0)

    @property
    def status(self) -> str | None:
        """The status of the author profile."""
        return self._profile.get("status")

    @property
    def subject_areas(self) -> list[Subjectarea] | None:
        """List of named tuples of subject areas in the form
        `(area, abbreviation, code)` of author's publication.
        """
        path = ['subject-areas', 'subject-area']
        areas = [Subjectarea(area=item['$'], code=int(item['@code']),
                      abbreviation=item['@abbrev'])
                 for item in chained_get(self._json, path, [])]
        return areas or None

    @property
    def surname(self) -> str | None:
        """Author's preferred surname."""
        return html_unescape(chained_get(self._profile, ['preferred-name', 'surname']))

    @property
    def url(self) -> str | None:
        """URL to the author's API page."""
        return chained_get(self._json, ['coredata', 'prism:url'])

    def __init__(self,
                 author_id: int | str,
                 refresh: bool | int = False,
                 view: str = "ENHANCED",
                 **kwds: str
                 ) -> None:
        """Interaction with the Author Retrieval API.

        :param author_id: The ID or the EID of the author.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param view: The view of the file that should be downloaded.  Allowed
                     values: `METRICS`, `LIGHT`, `STANDARD`, `ENHANCED`, `ENTITLED`, where `STANDARD`
                     includes all information of `LIGHT` view and `ENHANCED`
                     includes all information of any view.  For details see
                     https://dev.elsevier.com/sc_author_retrieval_views.html.
                     Note: Neither the `BASIC` nor the `DOCUMENTS` view are active,
                     although documented. `ENTITLED` only contains the `document_entitlement_status`.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/AuthorRetrievalAPI.wadl.

        Raises
        ------
        ValueError
            If any of the parameters `refresh` or `view` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/ENHANCED/{author_id}`,
        where `path` is specified in your configuration file, and `author_id`
        is stripped of an eventually leading `'9-s2.0-'`.
        """
        # Checks
        check_parameter_value(view, VIEWS['AuthorRetrieval'], "view")

        # Load json
        self._id = str(author_id).split('-')[-1]
        self._view = view
        self._refresh = refresh
        Retrieval.__init__(self, identifier=self._id, **kwds)

        if self._view in ('METRICS', 'LIGHT', 'STANDARD', 'ENHANCED'):
            # Parse json
            self._json = self._json['author-retrieval-response']
            try:
                self._json = self._json[0]
            except KeyError:  # Incomplete forward
                alias_json = listify(self._json['alias']['prism:url'])
                self._alias = [d['$'].split(':')[-1] for d in alias_json]
                alias_str = ', '.join(self._alias)
                text = f'Author profile with ID {author_id} has been merged and '\
                    f'the main profile is now one of {alias_str}.  Please update '\
                    'your records manually.  Functionality of this object is '\
                    'reduced.'
                warn(text, UserWarning)
            else:
                self._alias = None
        elif self._view == 'ENTITLED':
            self._alias = None
        self._profile = self._json.get("author-profile", {})

    def __str__(self):
        """Return a summary string."""
        if self._view in ('STANDARD', 'ENHANCED', 'LIGHT'):
            date = self.get_cache_file_mdate().split()[0]
            main_aff = self.affiliation_current[0]
            s = f"{self.indexed_name} from {main_aff.preferred_name} in "\
                f"{main_aff.country},\npublished {int(self.document_count):,} "\
                f"document(s) since {self.publication_range[0]} "\
                f"\nwhich were cited by {int(self.cited_by_count):,} author(s) in "\
                f"{int(self.citation_count):,} document(s) as of {date}"
        elif self._view == 'METRICS':
            s = f'Author with ID {self._id}\n'\
                f'published {int(self.document_count):,} document(s)\n'\
                f'which were cited by {int(self.cited_by_count):,} author(s) '\
                f'in {int(self.citation_count):,} document(s)'
        return s

    def get_coauthors(self) -> list[Coauthor] | None:
        """Retrieves basic information about co-authors as a list of
        namedtuples in the form
        `(surname, given_name, id, areas, affiliation_id, name, city, country)`,
        where areas is a list of subject area codes joined by `"; "`.
        Note: Method retrieves information via individual queries which will
        not be cached.  The Scopus API returns 160 coauthors at most.
        """
        SIZE = 25
        # Get number of authors to search for
        url = self.coauthor_link
        if not url:
            return None
        res = get_content(url, api="AuthorSearch")
        data = loads(res.text)['search-results']
        N = int(data.get('opensearch:totalResults', 0))
        # Store information in namedtuples
        coauthors = []
        # Iterate over search results in chunks of `SIZE` results
        count = SIZE
        start = 0
        while start < N:
            params = {'start': start, 'count': count, 'accept': 'json'}
            res = get_content(url, api="AuthorSearch", params=params)
            data = loads(res.text)['search-results'].get('entry', [])
            # Extract information for each coauthor
            for entry in data:
                aff = entry.get('affiliation-current', {})
                try:
                    areas = [a['$'] for a in entry.get('subject-area', [])]
                except TypeError:  # Only one subject area given
                    areas = [entry['subject-area']['$']]
                new = Coauthor(surname=entry['preferred-name']['surname'],
                    given_name=entry['preferred-name'].get('given-name'),
                    id=int(entry['dc:identifier'].split(':')[-1]),
                    areas='; '.join(areas), name=aff.get('affiliation-name'),
                    affiliation_id=aff.get('affiliation-id'),
                    city=aff.get('affiliation-city'),
                    country=aff.get('affiliation-country'))
                coauthors.append(new)
            start += SIZE
        return coauthors or None

    def get_documents(self,
                      subtypes: list[str] | None = None,
                      *args: str, **kwds: str
                      ) -> list | None:
        """Return list of the author's publications using a `ScopusSearch()`
        query, where publications may fit a specified set of document subtypes.

        :param subtypes: The type of documents that should be returned.
        :param args: Parameters to be passed on to `ScopusSearch()`.
        :param kwds: Parameters to be passed on to `ScopusSearch()`.

        Note: To update these results, use `refresh`; the class' `refresh`
        parameter is not used here.
        """
        s = ScopusSearch(f'AU-ID({self.identifier})', **kwds)
        if subtypes:
            return [p for p in s.results if p.subtype in subtypes]
        else:
            return s.results

    def get_document_eids(self,
                          *args: str, **kwds: str
                          ) -> list[str] | None:
        """Return list of EIDs of the author's publications using
        a ScopusSearch() query.

        :param args: Parameters to be passed on to `ScopusSearch()`.
        :param kwds: Parameters to be passed on to `ScopusSearch()`.

        Note: To update these results, use `refresh`; the class' `refresh`
        parameter is not used here.
        """
        s = ScopusSearch(f'AU-ID({self.identifier})', *args, **kwds)
        return s.get_eids()

    def estimate_uniqueness(self,
                            query: str | None = None,
                            *args: str,
                            **kwds: str
                            ) -> int:
        """Return the number of Scopus author profiles similar to this profile
        via calls with `AuthorSearch()`.

        :param query: The query string to perform to search for authors.  If
                      `None`, the query is of form `"AUTHLAST() AND AUTHFIRST()"`
                      with the corresponding information included.  Provided
                      queries may include `"SUBJAREA()" OR "AF-ID() AND
                      SUBJAREA()"`.  For details see
                      https://dev.elsevier.com/tips/AuthorSearchTips.htm.
        :param args: Parameters to be passed on to `AuthorSearch()`.
        :param kwds: Parameters to be passed on to `AuthorSearch()`.
        """
        if not query:
            query = f"AUTHLAST({self.surname}) AND AUTHFIRST({self.given_name})"
        s = AuthorSearch(query, *args, **kwds)
        return s.get_results_size()



"""Module with the class SubjectClassifications."""

from pybliometrics import scopus


class SubjectClassifications(scopus.SubjectClassifications):
    def __init__(self,
                 query: dict,
                 refresh: bool | int = False,
                 fields: list[str] | tuple[str, ...] | None = None,
                 **kwds: str
                 ) -> None:
        """Interaction with the ScienceDirect Subject Classifications API.

        :param query: Query parameters and corresponding fields. Allowed keys
                      `'code'`, `'abbrev'`, `'description'`, `'detail'`. For more
                      details on search fields please refer to the `documentation 
                      <https://dev.elsevier.com/documentation/SubjectClassificationsAPI.wadl#d1e199>`__.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param fields: The fields to return when calling search results.
                       Allowed values: `'code'`, `'abbrev'`, `'description'`,
                       `'detail'`.  For details see the `documentation 
                       <https://dev.elsevier.com/documentation/SubjectClassificationsAPI.wadl#d1e199>`__.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the `API specification 
                     <https://dev.elsevier.com/documentation/SubjectClassificationsAPI.wadl>`__.

        Raises
        ------
        TypeError
            If returned fields are not passed in an iterable container.

        ValueError
            If any of the parameters `fields`, `refresh` or `query` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{fname}`,
        where `path` is specified in your configuration file, and `fname` is
        the md5-hashed version of `query` dict turned into string in format
        of `'key=value'` delimited by `'&'`.
        """
        self.__class__.__name__ = 'ScDirSubjectClassifications'
        super().__init__(query=query, refresh=refresh, fields=fields, **kwds)



from typing import NamedTuple

from pybliometrics.superclasses import Search
from pybliometrics.utils import check_field_consistency, chained_get, \
    check_integrity, check_parameter_value, deduplicate, \
    make_search_summary, VIEWS


class Document(NamedTuple):
    """Named tuple representing a document from ScienceDirect Article Metadata API."""
    authorKeywords: str | None
    authors: str | None
    available_online_date: str | None
    first_author: str | None
    abstract_text: str | None
    doi: str | None
    title: str | None
    eid: str | None
    link: str | None
    openArchiveArticle: bool | None
    openaccess_status: str | None
    openaccessArticle: bool | None
    openaccessUserLicense: str | None
    pii: str | None
    aggregationType: str | None
    copyright: str | None
    coverDate: str | None
    coverDisplayDate: str | None
    edition: str | None
    endingPage: str | None
    isbn: str | None
    publicationName: str | None
    startingPage: str | None
    teaser: str | None
    api_link: str | None
    publicationType: str | None
    vor_available_online_date: str | None


class ArticleMetadata(Search):
    @property
    def results(self) -> list[Document] | None:
        """A list of namedtuples in the form `(authorKeywords authors available_online_date
        first_author abstract_text doi title eid link openArchiveArticle openaccess_status
        openaccessArticle openaccessUserLicense pii aggregationType copyright coverDate
        coverDisplayDate edition endingPage isbn publicationName startingPage teaser
        api_link publicationType vor_available_online_date)`.

        Field definitions correspond to the `Article Metadata Views
        <https://dev.elsevier.com/sd_article_meta_views.html>`__ and return the
        values as-is, except for `authors` which are joined on `";"`.

        Raises
        ------
        ValueError
            If the elements provided in `integrity_fields` do not match the
            actual field names (listed above).

        Notes
        -----
        The list of authors and the list of affiliations per author are
        deduplicated.
        """
        fields = 'authorKeywords authors available_online_date first_author abstract_text ' \
            'doi title eid link openArchiveArticle openaccess_status openaccessArticle '\
            'openaccessUserLicense pii aggregationType copyright coverDate coverDisplayDate '\
            'edition endingPage isbn publicationName startingPage teaser api_link publicationType '\
            'vor_available_online_date'
        check_field_consistency(self._integrity, fields)
        # Parse elements one-by-one
        out = []
        for item in self._json:
            # Get authors and create ";" separated string
            authors_list = [author.get('$') for author in chained_get(item, ['authors', 'author'], [])]
            authors_list = deduplicate(authors_list)
            authors = ';'.join(authors_list)
            first_author = item.get('dc:creator')[0].get('$')
            link = item.get('link')[0].get('@href')
            doi = item.get("prism:doi") or item.get("dc:identifier")[4:] if item.get("dc:identifier") else None
            new = Document(
                authorKeywords=item.get('authkeywords'),
                authors=authors,
                available_online_date=item.get('available-online-date'),
                first_author=first_author,
                abstract_text=item.get('dc:description'),
                doi=doi,
                title=item.get('dc:title'),
                eid=item.get('eid'),
                link=link,
                openArchiveArticle=item.get('openArchiveArticle'),
                openaccess_status=item.get('openaccess'),
                openaccessArticle=item.get('openaccessArticle'),
                openaccessUserLicense=item.get('openaccessUserLicense'),
                pii=item.get('pii'),
                aggregationType=item.get('prism:aggregationType'),
                copyright=item.get('prism:copyright'),
                coverDate=item.get('prism:coverDate'),
                coverDisplayDate=item.get('prism:coverDisplayDate'),
                edition=item.get('prism:edition'),
                endingPage=item.get('prism:endingPage'),
                isbn=item.get('prism:isbn'),
                publicationName=item.get('prism:publicationName'),
                startingPage=item.get('prism:startingPage'),
                teaser=item.get('prism:teaser'),
                api_link=item.get('prism:url'),
                publicationType=item.get('pubType'),
                vor_available_online_date=item.get('vor-available-online-date'),
            )
            out.append(new)
        check_integrity(out, self._integrity, self._action)
        return out or None

    def __init__(self,
                 query: str,
                 refresh: bool | int = False,
                 view: str | None = None,
                 verbose: bool = False,
                 download: bool = True,
                 integrity_fields: list[str] | tuple[str, ...] | None = None,
                 integrity_action: str = "raise",
                 subscriber: bool = True,
                 **kwds: str
                 ) -> None:
        """Interaction with the ScienceDirect Article Metadata API.

        :param query: A string of the query as used in the `Advanced Search <https://dev.elsevier.com/tecdoc_sdsearch_migration.html>`__.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param view: Which view to use for the query, see `the documentation <https://dev.elsevier.com/sd_article_meta_views.html>`__.
                     Allowed values: `STANDARD`, `COMPLETE`.  If `None`, defaults to
                     `COMPLETE` if `subscriber=True` and to `STANDARD` if
                     `subscriber=False`.
        :param verbose: Whether to print a download progress bar.
        :param download: Whether to download results (if they have not been
                         cached).
        :param integrity_fields: A list or tuple with the names of fields whose completeness should
                                 be checked.  `ArticleMetadata` will perform the
                                 action specified in `integrity_action` if
                                 elements in these fields are missing.  This
                                 helps to avoid idiosynchratically missing
                                 elements that should always be present
                                 (e.g., EID or source ID).
        :param integrity_action: What to do in case integrity of provided fields
                                 cannot be verified.  Possible actions:
                                 - `"raise"`: Raise an `AttributeError`
                                 - `"warn"`: Raise a `UserWarning`
        :param subscriber: Whether you access ScienceDirect with a subscription or not.
                           For subscribers, ScienceDirect's cursor navigation will be
                           used.  Sets the number of entries in each query
                           iteration to the maximum number allowed by the
                           corresponding view.
        :param unescape: Convert named and numeric characters in the `results` to
                        their corresponding Unicode characters.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the `API specification <https://dev.elsevier.com/documentation/ArticleMetadataAPI.wadl>`__.

        Raises
        ------
        ScopusQueryError
            For non-subscribers, if the number of search results exceeds 5000.

        ValueError
            If any of the parameters `integrity_action`, `refresh` or `view`
            is not one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{view}/{fname}`,
        where `path` is specified in your configuration file and `fname` is
        the md5-hashed version of `query`.
        """
        # Check view or set to default
        if view:
            check_parameter_value(view, VIEWS['ArticleMetadata'], "view")
        else:
            view = "COMPLETE" if subscriber else "STANDARD"

        allowed = ("warn", "raise")
        check_parameter_value(integrity_action, allowed, "integrity_action")

        # Query
        self._action = integrity_action
        self._integrity = integrity_fields or []
        self._refresh = refresh
        self._query = query
        self._view = view
        Search.__init__(self, query=query, download=download, verbose=verbose, **kwds)

    def __str__(self):
        """Print a summary string."""
        return make_search_summary(self, "document", self.get_eids())

    def get_eids(self):
        """EIDs of retrieved documents."""
        return [d['eid'] for d in self._json]
